import logging

from .coroutine import circuit_breaker as coroutine_circuit_breaker

logger = logging.getLogger(__name__)


class circuit_breaker(coroutine_circuit_breaker):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def increment(self):
        return await super().increment()

    async def open_circuit(self):
        await super().open_circuit()

    async def _check_circuit(self):
        await super()._check_circuit()



import abc
import logging

logger = logging.getLogger(__name__)


class BaseCircuitBreaker(metaclass=abc.ABCMeta):
    """
    This class provides the basic logics to make the circuit breaker
    mechanism work as expected. Every logic must be implemented here,
    the other modules inside this package are strongly dependent on
    which kind of asynchronous model your
    code is running with (async/await or coroutines).
    """

    def __init__(
        self,
        storage,
        failure_key,
        max_failures,
        max_failure_exception,
        max_failure_timeout=None,
        circuit_timeout=None,
        catch_exceptions=None
    ):
        self.storage = storage
        self.failure_key = failure_key
        self.max_failure_timeout = max_failure_timeout
        self.circuit_timeout = circuit_timeout
        self.circuit_key = 'circuit_{}'.format(failure_key)
        self.max_failure_exception = max_failure_exception
        self.catch_exceptions = catch_exceptions or (Exception,)
        self.max_failures = max_failures

    @abc.abstractmethod
    def increment(self):
        """
        This method demands that the implementation is responsible for
        getting a storage key from the storage engine.
        """

    @property
    @abc.abstractmethod
    def is_circuit_open(self):
        """
        This method must implement a call to the storage engine in order
        to retrieve the current circuit state, based on this instance's key.
        """

    @abc.abstractmethod
    def open_circuit(self):
        """
        This method must "open the circuit" by setting this instance's key
        on the storage engine, and also, setting a TTL to it.
        """

    def _is_catchable(self, exception):
        is_catchable = any(
            exc in self.catch_exceptions
            for exc in [type(exception), exception]
        )

        logger.debug('Testing if {} is catcheable:{}'.format(
            type(exception),
            is_catchable
        ))

        return is_catchable

    def _exceeded_max_failures(self, total_failures):
        return total_failures >= self.max_failures

    def _raise_openess(self):
        raise self.max_failure_exception



import logging

from .base import BaseCircuitBreaker

logger = logging.getLogger(__name__)


class CircuitBreaker(BaseCircuitBreaker):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def increment(self):
        """
        This method demands that the implementation is responsible for
        getting a storage key from the storage engine.
        """
        total = self.storage.increment(self.failure_key)

        logger.info(
            'Increase failure for: {key} - '
            'max failures {max_failures} - '
            'total {total}'.format(
                key=self.failure_key,
                max_failures=self.max_failures,
                total=total
            )
        )

        return int(total or 0)

    @property
    def is_circuit_open(self):
        return self.storage.get(self.circuit_key) or False

    def open_circuit(self):
        self.storage.set(
            self.circuit_key,
            1,
            self.circuit_timeout
        )

    def __enter__(self):
        self._check_circuit()

        return self

    def _check_circuit(self):
        if self.is_circuit_open:
            self._raise_openess()

    def __exit__(self, exc_type, exc_value, traceback):
        if self._is_catchable(exc_type):
            self._check_circuit()

            total_failures = self.increment()

            if self._exceeded_max_failures(total_failures):
                self.open_circuit()

                logger.info(
                    'Max failures exceeded by: {}'.format(
                        self.failure_key
                    )
                )

                raise self.max_failure_exception



import abc

from werkzeug.contrib.cache import SimpleCache


class CircuitBreakerBaseStorage(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get(self, key):
        """
        This method must implement a GET call to a stored value.
        It should return an int value that represents how many times
        this single key has been set.
        """

    @abc.abstractmethod
    def increment(self, key):
        """
        This method must increment 1 int value based in a key stored value.
        It should return an int value that represents how many times
        this single key has been set, after it was incremented by its request.
        """

    @abc.abstractmethod
    def set(self, key, value, timeout):
        """
        This method must add a key with an int value and set a ttl to it
        """


class MemoryStorage(CircuitBreakerBaseStorage):

    def __init__(self):
        self._cache = SimpleCache()

    def get(self, key):
        return self._cache.get(key)

    def increment(self, key):
        return self._cache.inc(key)

    def set(self, key, value, timeout):
        self._cache.set(key, value, timeout)



import asyncio
import logging
from functools import wraps

from .base import BaseCircuitBreaker

logger = logging.getLogger(__name__)


class circuit_breaker(BaseCircuitBreaker):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @asyncio.coroutine
    def increment(self):
        """
        This method demands that the implementation is responsible for
        getting a storage key from the storage engine.
        """

        total = yield from self.storage.increment(self.failure_key, 1)
        logger.info(
            'Increase failure for: {key} - '
            'max failures {max_failures} - '
            'total {total}'.format(
                key=self.failure_key,
                max_failures=self.max_failures,
                total=total
            )
        )

        return int(total or 0)

    @property
    @asyncio.coroutine
    def is_circuit_open(self):
        return (yield from self.storage.get(self.circuit_key)) or False

    @asyncio.coroutine
    def open_circuit(self):
        yield from self.storage.set(self.circuit_key, 1, self.circuit_timeout)
        yield from self.storage.delete(self.failure_key)

        logger.critical(
            'Open circuit for {failure_key} {cicuit_storage_key}'.format(
                failure_key=self.failure_key,
                cicuit_storage_key=self.circuit_key
            )
        )

    @asyncio.coroutine
    def _check_circuit(self):
        is_open = yield from self.is_circuit_open
        if is_open:
            self._raise_openess()

    def __call__(self, method):
        @asyncio.coroutine
        @wraps(method)
        def wrapper(*args, **kwargs):
            yield from self._check_circuit()

            try:
                return (yield from method(*args, **kwargs))
            except Exception as e:
                if self._is_catchable(e):
                    yield from self._check_circuit()

                    total_failures = yield from self.increment()

                    if self._exceeded_max_failures(total_failures):
                        yield from self.open_circuit()

                        logger.info(
                            'Max failures exceeded by: {}'.format(
                                self.failure_key
                            )
                        )

                        raise self.max_failure_exception
                else:
                    raise e

        return wrapper



import os

from setuptools import find_packages, setup


def read(fname):
    path = os.path.join(os.path.dirname(__file__), fname)
    with open(path) as f:
        return f.read()


setup(
    name='luizalabs-asyncio-toolkit',
    version='0.2.2',
    description=(
        'The LuizaLabs set of tools '
        'to ease asyncio development'
    ),
    long_description=read('README.rst'),
    author='Luizalabs',
    author_email='pypi@luizalabs.com',
    url='https://github.com/luizalabs/asyncio-toolkit',
    keywords='asyncio tools utils circuit breaker',
    install_requires=[],
    packages=find_packages(exclude=[
        'tests*'
    ]),
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
    ]
)
